from Packages import *
class Sk_Kmeans:
    """K-Means clustering for Samples selection.

    Returns:
        inertia_ (pd.DataFrame): DataFrame with ...
        x (pd.DataFrame): Initial data
        clu (pd.DataFrame): Cluster name for each sample
        model.cluster_centers_ (pd.DataFrame): Coordinates of the center of each cluster
    """
    def __init__(self, x, max_clusters):
        """Initiate the KMeans class.

        Args:
            x (pd.DataFrame): the original reduced data to cluster
            max_cluster (Int): the max number of desired clusters.
        """
        self.x = x
        self.max_clusters = max_clusters

        self.inertia = pd.DataFrame()
        for i in range(1, max_clusters+1):
            model = KMeans(n_clusters = i, init = 'k-means++', random_state = 42)
            model.fit(x)
            self.inertia[f'{i}_clust']= [model.inertia_]
        self.inertia.index = ['inertia']

    @property
    def inertia_(self):
        return self.inertia
    
    @property
    def suggested_n_clusters_(self):
        idxidx = []
        values = []

        s = self.inertia.to_numpy().ravel()
        for i in range(self.max_clusters-1):
            idxidx.append(f'{i+1}_clust')
            values.append((s[i] - s[i+1])*100 / s[i])

        id = np.max(np.where(np.array(values) > 10))+2
        return id
    
    def fit_optimal(self, nclusters):
        model = KMeans(n_clusters = nclusters, init = 'k-means++', random_state = 42)
        model.fit(self.x)
        yp = model.predict(self.x)+1
        clu = [f'cluster#{i}' for i in yp]

        return self.x, clu, model.cluster_centers_